package markov;

import com.google.common.base.Stopwatch;
import markov.huffman.ByteHuffmanCodeBuilder;
import markov.huffman.HuffmanCode;
import markov.huffman.HuffmanStore;
import markov.stuff.CountMap;
import markov.stuff.Inspector;
import markov.stuff.SimpleCountMap;
import markov.stuff.Utils;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.IntSummaryStatistics;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;

public class Mail {

    private static final String intsFileLocation = System.getProperties().getProperty(
            "huffmanints.file", "classpath:huffmanInts");
    private static final String bytesFileLocation = System.getProperties().getProperty(
            "huffmanbytes.file", "classpath:huffmanBytes");

        private static final int iterations = 1_000_000;
//    private static final int iterations = 10_000;

    public static void main(String[] args)
    throws IOException, ClassNotFoundException, NoSuchFieldException, IllegalAccessException {
        System.in.read();
        new Mail().timeCreation();
    }

    private void timeCreation() {
        final int maxPrefix = 7;
        Map<Integer, Data> markovData = Utils.createDataMap(maxPrefix);
        System.out.println("initializing...");
        final Builder[] builder = new Builder[maxPrefix + 1];
        IntStream.range(1, maxPrefix + 1).sequential().forEach(i -> {
            Utils.maybePrintPercentages("init trainer", i, maxPrefix);
            Data data = markovData.get(i);
            builder[i] = new Builder(i, data);
        });

        System.out.println("creating...");
        Stopwatch s = Stopwatch.createStarted();
        for (int i = 0; i < iterations; i++) {
            Utils.maybePrintPercentages(i, iterations);
            IntStream.range(1, maxPrefix + 1).parallel().forEach(j -> builder[j].random());
        }
        s.stop();
        System.out.println("took: " + s.toString());
    }

    private void createHuffmanMaps() {
        final int maxPrefix = 7;
        Map<Integer, Data> markovData = Utils.createDataMap(maxPrefix);
        System.out.println("initializing...");
        final ShortenerByteHUffmanTrainerImpl[] trainer =
                new ShortenerByteHUffmanTrainerImpl[maxPrefix + 1];
        final Builder[] builder = new Builder[maxPrefix + 1];
        IntStream.range(1, maxPrefix + 1).sequential().forEach(i -> {
            Utils.maybePrintPercentages("init trainer", i, maxPrefix);
            Data data = markovData.get(i);
            trainer[i] = new ShortenerByteHUffmanTrainerImpl(data);
            builder[i] = new Builder(i, data);
        });

        System.out.println("creating...");
        Stopwatch s = Stopwatch.createStarted();
        for (int i = 0; i < iterations; i++) {
            Utils.maybePrintPercentages(i, iterations);
            IntStream.range(1, maxPrefix + 1).parallel().forEach(
                    j -> trainer[j].getId(builder[j].random()));
//            for (int j = 1; j <= maxPrefix; j++) {
//                trainer[j].getId(builder[j].random());
//            }
        }
        s.stop();
        System.out.println("took: " + s.toString());
        System.out.println("saving...");
        Map<Integer, CountMap<Integer>> intData = new HashMap<>();
        Map<Integer, CountMap<Byte>> byteData = new HashMap<>();
        IntStream.range(1, maxPrefix + 1).forEach(i -> {
            intData.put(i + 1, trainer[i].getIntsMap());
            byteData.put(i, trainer[i].getByteMap());
        });
        System.out.println("writing to file");
        saveHuffmanMaps(intData, byteData);

    }

    @SuppressWarnings("unchecked")
    private Map<Integer, CountMap<Integer>> readIntMap() {
        HuffmanStore<Integer> store = readHUffmanMap(intsFileLocation);
        return store.getInternal();
    }

    @SuppressWarnings("unchecked")
    private Map<Integer, CountMap<Byte>> readByteMap() {
        HuffmanStore<Byte> store = readHUffmanMap(bytesFileLocation);
        return store.getInternal();
    }

    // we controll whats in the file
    @SuppressWarnings("unchecked")
    private <X extends Serializable> HuffmanStore<X> readHUffmanMap(String location) {
        File file = Utils.getFile(location);
        System.out.println("reading from: " + file.getAbsolutePath());
        try (ObjectInputStream in = new ObjectInputStream(new FileInputStream(file))) {

            return (HuffmanStore<X>) in.readObject();
        } catch (IOException | ClassNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    private void saveHuffmanMaps(
            Map<Integer, CountMap<Integer>> intCounts, Map<Integer, CountMap<Byte>> byteCounts) {

        writeCounts(new HuffmanStore<>(intCounts), intsFileLocation);
        writeCounts(new HuffmanStore<>(byteCounts), bytesFileLocation);
    }

    private void writeCounts(HuffmanStore<?> counts, String intsFileLocation) {
        File file = Utils.getFile(intsFileLocation);
        try (ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(file))) {
            out.writeObject(counts);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private void printPrefixes() throws NoSuchFieldException, IllegalAccessException {
        Map<Integer, Data> data = Utils.createDataMap(4);
        System.out.println("\n\n\n1:");
        Inspector.extractMap(data.get(1)).keySet().stream().sorted().limit(20).forEach(
                System.out::println);
        System.out.println("\n\n\n2:");
        Inspector.extractMap(data.get(2)).keySet().stream().sorted().limit(20).forEach(
                System.out::println);
        System.out.println("\n\n\n3:");
        Inspector.extractMap(data.get(3)).keySet().stream().sorted().limit(20).forEach(
                System.out::println);
        System.out.println("\n\n\n4:");
        Inspector.extractMap(data.get(4)).keySet().stream().sorted().limit(20).forEach(
                System.out::println);
    }


    private void printDataHashStats() throws NoSuchFieldException, IllegalAccessException {
        Map<Integer, Data> data = Utils.createDataMap(7);
        Inspector.printDataHashStats(data);
    }


    private static class ShortenerStats {
        private final Shortener shortener;
        private final List<Integer> stats = new ArrayList<>(iterations);

        private ShortenerStats(Shortener shortener) {
            this.shortener = shortener;
        }

        public void record(Sentence sentence) {
            stats.add(shortener.getId(sentence).length());
        }

        public void printStats() {
            System.out.println(shortener.getClass().getName());
            System.out.println("distinct: " + stats.stream().distinct().count());
            IntSummaryStatistics statistics = stats.stream()
                                                   .mapToInt(Integer::intValue)
                                                   .summaryStatistics();
            System.out.println(String.format("avg, min, high: %f, %d, %d", statistics.getAverage(),
                                             statistics.getMin(), statistics.getMax()));
            System.out.println("SD: " + Math.sqrt(stats.stream()
                                                       .mapToDouble(e -> (double) (Math.pow(
                                                               e - statistics.getAverage(), 2)))
                                                       .sum() / iterations));
        }
    }

    public void printShorterStats() throws IOException, ClassNotFoundException {
        int prefixLength = 4;
        HuffmanCode<Integer, List<Boolean>, Boolean, SimpleCountMap.MutableInt> intCode =
                new ByteHuffmanCodeBuilder<Integer, SimpleCountMap.MutableInt>().generateCode(
                        readIntMap().get(prefixLength).asMap());
        HuffmanCode<Byte, List<Boolean>, Boolean, SimpleCountMap.MutableInt> byteCode =
                new ByteHuffmanCodeBuilder<Byte, SimpleCountMap.MutableInt>().generateCode(
                        readByteMap().get(prefixLength).asMap());

        Map<Integer, Data> dataMap = Utils.createDataMap(prefixLength);
        Builder b = new Builder(prefixLength, dataMap.get(prefixLength));
        List<ShortenerStats> shorter = Arrays.asList(
                new ShortenerStats(new ShortenerSimpleImpl(dataMap.get(prefixLength))),
                new ShortenerStats(new ShortenerIntHuffmanImpl(dataMap.get(prefixLength), intCode)),
                new ShortenerStats(
                        new ShortenerByteHuffmanImpl(dataMap.get(prefixLength), byteCode)));

        System.out.println("creating...");
        for (int i = 0; i < iterations; i++) {
            Utils.maybePrintPercentages(i, iterations);
            Sentence sentence = b.random();
            for (ShortenerStats s : shorter) {
                s.record(sentence);
            }
        }
        System.out.println("printing stats:");
        for (ShortenerStats s : shorter) {
            s.printStats();
        }
    }


}
