package markov.huffman;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * Used to build a HUffman Code instance to code a given ContentType into a SequenceType,
 * and decode a sequence of SequenceTypes into a sequence of ContentTypes,
 * using the given Frequency type to determine the  distribution.
 *
 * @param <ContentType>
 * @param <SequenceType>
 * @param <GlyphType>
 * @param <FrequencyType>
 */
public class HuffmanCodeBuilder<ContentType, SequenceType, GlyphType, FrequencyType extends FrequenceType<FrequencyType>> {

    private final Supplier<SequenceType> rootCodeSupplier;
    private final Supplier<GlyphType> leftGlyph;
    private final Supplier<GlyphType> rightGlyph;
    private final BiFunction<SequenceType, GlyphType, SequenceType> combiner;
    private final Function<SequenceType, Iterator<GlyphType>> splitter;


    private final BiFunction<HuffmanNode<ContentType, SequenceType, FrequencyType>,
            GlyphType, HuffmanTree<ContentType, SequenceType, FrequencyType>> leftRightDecider;

    private HuffmanNode<ContentType, SequenceType, FrequencyType> tree;
    private final Map<ContentType, SequenceType> codes = new HashMap<>();


    public HuffmanCodeBuilder(Supplier<SequenceType> rootCodeSupplier, Supplier<GlyphType> leftGlyph, Supplier<GlyphType> rightGlyph, BiFunction<SequenceType, GlyphType, SequenceType> combiner, Function<SequenceType, Iterator<GlyphType>> splitter, BiFunction<HuffmanNode<ContentType, SequenceType, FrequencyType>, GlyphType, HuffmanTree<ContentType, SequenceType, FrequencyType>> leftRightDecider) {
        this.rootCodeSupplier = rootCodeSupplier;
        this.leftGlyph = leftGlyph;
        this.rightGlyph = rightGlyph;
        this.combiner = combiner;
        this.splitter = splitter;
        this.leftRightDecider = leftRightDecider;
    }


    public HuffmanCode<ContentType, SequenceType, GlyphType, FrequencyType> generateCode(Map<ContentType, FrequencyType> frequencies) {
        tree = buildTree(frequencies);
        tree.setCode(rootCodeSupplier.get());
        generateCodes(tree);
        return new HuffmanCode<>(codes, tree, splitter, leftRightDecider);
    }

    private HuffmanNode<ContentType, SequenceType, FrequencyType> buildTree(Map<ContentType, FrequencyType> frequencies) {
        PriorityQueue<HuffmanTree<ContentType, SequenceType, FrequencyType>> trees = new PriorityQueue<>();
        for (Map.Entry<ContentType, FrequencyType> e : frequencies.entrySet()) {
            if (e.getValue().isGreaterZero())
                trees.offer(new HuffmanLeaf<>(e.getValue(), e.getKey()));
        }

        while (trees.size() > 1) {
            HuffmanTree a = trees.poll();
            HuffmanTree b = trees.poll();
            trees.offer(new HuffmanNode<ContentType, SequenceType, FrequencyType>(a, b));
        }
        return (HuffmanNode<ContentType, SequenceType, FrequencyType>) trees.poll();
    }

    private void generateCodes(HuffmanTree<ContentType, SequenceType, FrequencyType> child) {
        if (child == null)
            throw new IllegalArgumentException(new NullPointerException());
        if (child instanceof HuffmanLeaf) {
            HuffmanLeaf<ContentType, SequenceType, FrequencyType> leaf = (HuffmanLeaf<ContentType, SequenceType, FrequencyType>) child;
            codes.put(leaf.getValue(), leaf.getCode());
        } else if (child instanceof HuffmanNode) {
            HuffmanNode<ContentType, SequenceType, FrequencyType> node = (HuffmanNode<ContentType, SequenceType, FrequencyType>) child;

            // traverse left
            HuffmanTree<ContentType, SequenceType, FrequencyType> left = node.getLeft();
            SequenceType leftCode = combiner.apply(child.getCode(), leftGlyph.get());
            left.setCode(leftCode);
            generateCodes(left);
            // traverse right
            HuffmanTree<ContentType, SequenceType, FrequencyType> right = node.getRight();
            SequenceType rightCode = combiner.apply(child.getCode(), rightGlyph.get());
            right.setCode(rightCode);
            generateCodes(right);
        }
    }
}
